from datetime import date
import os
import shutil

import toml

import audeer


config = toml.load(audeer.path("..", "pyproject.toml"))


# Project -----------------------------------------------------------------
project = config["project"]["name"]
copyright = f"2019-{date.today().year} audEERING GmbH"
author = ", ".join(author["name"] for author in config["project"]["authors"])
version = audeer.git_repo_version()
title = "Documentation"


# General -----------------------------------------------------------------
master_doc = "index"
source_suffix = ".rst"
exclude_patterns = [
    "build",
    "tests",
    "Thumbs.db",
    ".DS_Store",
    "api-src",
]
templates_path = ["_templates"]
pygments_style = None
extensions = [
    "sphinxcontrib.bibtex",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",  # support for Google-style docstrings
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinxcontrib.katex",  # has to be before jupyter_sphinx
]

napoleon_use_ivar = True  # List of class attributes
autodoc_inherit_docstrings = False  # disable docstring inheritance
bibtex_bibfiles = ["refs.bib"]
# Don't check for DOIs as they will always work
# and tend to fail the test
linkcheck_ignore = [
    "https://doi.org/",
]

# Disable auto-generation of TOC entries in the API
# https://github.com/sphinx-doc/sphinx/issues/6316
toc_object_entries = False


# HTML --------------------------------------------------------------------
html_theme = "sphinx_audeering_theme"
html_theme_options = {
    "display_version": True,
    "logo_only": False,
    "footer_links": False,
}
html_context = {
    "display_github": True,
}
html_title = title


# Copy API (sub-)module RST files to docs/api/ folder ---------------------
audeer.rmdir("api")
audeer.mkdir("api")
api_src_files = audeer.list_file_names("api-src")
api_dst_files = [
    audeer.path("api", os.path.basename(src_file)) for src_file in api_src_files
]
for src_file, dst_file in zip(api_src_files, api_dst_files):
    shutil.copyfile(src_file, dst_file)



import typing

import numpy as np


def assert_equal_length(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
):
    r"""Assert truth and prediction have equal length."""
    if len(truth) != len(prediction):
        raise ValueError(
            f"Truth and prediction differ in length: "
            f"{len(truth)} != {len(prediction)}."
        )


def infer_labels(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
) -> typing.List[typing.Any]:
    r"""Infer labels from truth and prediction.

    It gathers all labels that are present
    in the truth and prediction values.

    Args:
        truth: ground truth labels
        prediction: predicted labels

    Returns:
        labels in sorted order

    """
    return sorted(list(set(truth) | set(prediction)))


def scores_per_subgroup_and_class(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    protected_variable: typing.Sequence[typing.Any],
    metric: typing.Callable[
        [
            typing.Sequence[typing.Any],
            typing.Sequence[typing.Any],
            typing.Optional[typing.Sequence[str]],
            float,
        ],
        typing.Dict[str, float],
    ],
    labels: typing.Sequence[typing.Any],
    subgroups: typing.Sequence[typing.Any],
    zero_division: float,
) -> typing.Dict[typing.Hashable, typing.Dict]:
    r"""Compute scores per class for each subgroup based on metric.

    Args:
        truth: ground truth classes
        prediction: predicted classes
        protected_variable: manifestations of protected variable such as
            subgroups "male" and "female" of variable "sex"
        metric: metric to measure performance
        labels: included labels in preferred ordering
        subgroups: included subgroups in preferred ordering
        zero_division: set the value to return when there is a zero division

    Returns:
        scores per class for each subgroup

    Raises:
        ValueError: if `subgroups` contains values not contained in
            `protected_variable`

    Examples:
        >>> import audmetric
        >>> scores_per_subgroup_and_class(
        ...     [1, 1],
        ...     [0, 1],
        ...     ["male", "female"],
        ...     audmetric.recall_per_class,
        ...     [0, 1],
        ...     ["male", "female"],
        ...     0.0,
        ... )
        {'male': {0: 0.0, 1: 0.0}, 'female': {0: 0.0, 1: 1.0}}
        >>> scores_per_subgroup_and_class(
        ...     [1, 1],
        ...     [0, 1],
        ...     ["male", "female"],
        ...     audmetric.precision_per_class,
        ...     [0, 1],
        ...     ["male", "female"],
        ...     zero_division=np.nan,
        ... )
        {'male': {0: 0.0, 1: nan}, 'female': {0: nan, 1: 1.0}}

    """  # noqa: E501
    if set(subgroups) - set(protected_variable):
        raise ValueError(
            f"`subgroups` contains manifestations of the protected "
            f"variable which are not contained in `protected_variable`: "
            f"{set(subgroups) - set(protected_variable)}"
        )

    truth = np.array(truth)
    prediction = np.array(prediction)
    protected_variable = np.array(protected_variable)

    score = {}
    for subgroup in subgroups:
        mask = protected_variable == subgroup
        score[subgroup] = metric(
            truth[mask],
            prediction[mask],
            labels,
            zero_division=zero_division,
        )
    return score



import subprocess
import tempfile

import numpy as np

import audeer

import audmetric


def truth(samples=1000):
    r"""Generate truth values for linkability test.

    The truth values are in the form:
    ``[0, 1, 0, 1, ..., 0, 1]``

    """
    return [0, 1] * samples


def prediction(range1, range2, samples=1000, random_state=1):
    r"""Generate prediction values for linkability test.

    The values are returned as pairs
    drawn from uniform distribution
    given by ``range1`` and ``range2``, e.g.
    ``[0.12, 0.84, 0.07, 0.81, ..., 0.13, 0.84]``

    """
    np.random.seed(random_state)
    return audeer.flatten_list(
        [
            [np.random.uniform(*range1), np.random.uniform(*range2)]
            for _ in range(samples)
        ]
    )


test_cases = [
    # truth, prediction, omega
    (
        truth(),
        prediction((0.0, 0.2), (0.8, 1.0)),
        1,
    ),
    (
        truth(),
        prediction((0.0, 0.2), (0.8, 1.0)),
        1 / 9.0,
    ),
    (
        truth(),
        prediction((0.1, 0.3), (0.7, 0.9)),
        1,
    ),
    (
        truth(),
        prediction((0.2, 0.4), (0.6, 0.8)),
        1,
    ),
    (
        truth(),
        prediction((0.3, 0.5), (0.5, 0.7)),
        1,
    ),
    (
        truth(),
        prediction((0.4, 0.6), (0.5, 0.7)),
        1,
    ),
    (
        truth(),
        prediction((0.3, 0.5), (0.4, 0.6)),
        1,
    ),
    (
        truth(),
        prediction((0.4, 0.6), (0.4, 0.6)),
        1,
    ),
    (
        truth(),
        prediction((0.5, 0.7), (0.3, 0.5)),
        1,
    ),
    (
        truth(),
        prediction((0.3, 0.6), (0.5, 0.8)),
        1,
    ),
    (
        truth(),
        prediction((0.2, 0.6), (0.5, 0.9)),
        1,
    ),
    (
        truth(),
        prediction((0.1, 0.6), (0.5, 1.0)),
        1,
    ),
    (
        truth(),
        prediction((0.4, 0.6), (0.5, 0.7)),
        1,
    ),
    (
        truth(),
        prediction((0.0, 1.0), (0.0, 1.0)),
        1.0,
    ),
    (
        truth(),
        prediction((0.0, 1.0), (0.0, 1.0)),
        1 / 2.0,
    ),
    (
        truth(),
        prediction((0.0, 1.0), (0.0, 1.0)),
        2.0 / 1,
    ),
]

results = []
for n, (truth, prediction, omega) in enumerate(test_cases):
    mated_scores, nonmated_scores = audmetric.core.api._matching_scores(
        truth,
        prediction,
    )
    with tempfile.TemporaryDirectory() as tmp:
        score_file = audeer.path(tmp, "score")
        with open(score_file, "w") as fp:
            for score in mated_scores:
                fp.write(f"{score} 1\n")
            for score in nonmated_scores:
                fp.write(f"{score} 0\n")
        shell_command = [
            "python",
            "anonymization_metrics/compute_metrics.py",
            "-s",
            f"{score_file}",
            "--omega",
            f"{omega}",
        ]
        out = subprocess.check_output(shell_command, stderr=subprocess.STDOUT)
        # The above Python command returns a string containing:
        # id,matedMean,nonMatedMean,matedStd,nonMatedStd,linkability,cllr,cmin,eer  # noqa: E501
        # which means we can access the linkability by -4 from the end
        linkability = out.split()[0].decode("utf-8").split(",")[-4]
        results.append((omega, linkability))

# Store results
with open("results.txt", "w") as fp:
    fp.write("testcase,omega,linkability\n")
    for n, (omega, linkability) in enumerate(results):
        fp.write(f"{n},{omega},{linkability}\n")



import numpy as np
import pytest

import audmetric


@pytest.mark.parametrize(
    "truth,prediction,protected_variable,metric,labels,subgroups,"
    "zero_division,expected",
    [
        pytest.param(
            [],
            [],
            [],
            None,
            [],
            [0],
            None,
            {},
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        ([], [], [], audmetric.recall_per_class, [], [], 0.0, {}),
        (
            [1],
            [0],
            [0],
            audmetric.recall_per_class,
            [0, 1],
            [0],
            0.0,
            {0: {0: 0.0, 1: 0.0}},
        ),
        (
            [1],
            [0],
            [0],
            audmetric.precision_per_class,
            [0, 1],
            [0],
            0.0,
            {0: {0: 0.0, 1: 0.0}},
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            audmetric.recall_per_class,
            [0, 1],
            [0, 1],
            np.nan,
            {0: {0: np.nan, 1: 0.0}, 1: {0: np.nan, 1: 1.0}},
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            audmetric.recall_per_class,
            [1],
            [0, 1],
            np.nan,
            {0: {1: np.nan}, 1: {1: 1.0}},
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            audmetric.precision_per_class,
            [0, 1],
            [0, 1],
            np.nan,
            {0: {0: 0.0, 1: np.nan}, 1: {0: np.nan, 1: 1.0}},
        ),
    ],
)
def test_scores_per_subgroup_and_class(
    truth,
    prediction,
    protected_variable,
    metric,
    labels,
    subgroups,
    zero_division,
    expected,
):
    np.testing.assert_equal(
        audmetric.core.utils.scores_per_subgroup_and_class(
            truth,
            prediction,
            protected_variable,
            metric,
            labels=labels,
            subgroups=subgroups,
            zero_division=zero_division,
        ),
        expected,
    )



import numpy as np
import pandas as pd
import pytest
import sklearn.metrics

import audmetric


@pytest.mark.parametrize(
    "truth,prediction,labels,to_string",
    [
        (
            np.random.randint(0, 10, size=5),
            np.random.randint(0, 10, size=5),
            None,
            False,
        ),
        (
            pd.Series(np.random.randint(0, 10, size=5)).astype("Int64"),
            pd.Series(np.random.randint(0, 10, size=5)).astype("Int64"),
            None,
            False,
        ),
        (
            np.random.randint(0, 10, size=1),
            np.random.randint(0, 10, size=1),
            list(range(1, 10)),
            False,
        ),
        (
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=10),
            list(range(1, 10)),
            False,
        ),
        (
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=10),
            None,
            True,
        ),
        (
            np.array([]),
            np.array([]),
            None,
            False,
        ),
        (
            np.zeros(10),
            np.zeros(10),
            None,
            False,
        ),
        (
            np.arange(10),
            np.arange(10),
            list(range(1, 10)),
            False,
        ),
        (
            np.arange(10),
            np.arange(1, 11),
            list(range(1, 10)),
            False,
        ),
        (np.arange(5), np.array([1, 2, 3, 4, 6]), list(range(5)), False),
    ],
)
def test_accuracy(truth, prediction, labels, to_string):
    if to_string:
        truth = [str(w) for w in truth]
        prediction = [str(w) for w in prediction]

    if len(prediction) == 0:
        accuracy = np.nan
    else:
        if labels:
            mask = np.nonzero(
                np.logical_and(np.isin(truth, labels), np.isin(prediction, labels))
            )
            truth = truth[mask]
            prediction = prediction[mask]
        accuracy = sklearn.metrics.accuracy_score(
            list(truth),
            list(prediction),
        )

    np.testing.assert_almost_equal(
        audmetric.accuracy(truth, prediction, labels=labels),
        accuracy,
    )


@pytest.mark.parametrize(
    "truth, prediction, expected_fmr, expected_fnmr, expected_thresholds",
    [
        (
            [1, 1, 0, 0],
            [1, 1, 0, 0],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
        ),
        (
            [True, True, False, False],
            [True, True, False, False],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
        ),
        (
            [True, True, False, False],
            [1, 1, 0, 0],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
        ),
        (  # similarities > 1 or < 0
            [True, True, False, False],
            [1.5, 1.5, -1.0, -1.0],
            [1.0, 0.0],
            [0.0, 0.0],
            [-1.0, 1.5],
        ),
        (  # unordered
            [1, 0, 1, 1],
            [0.5, 0.4, 0.8, 0.2],
            [1.0, 1.0, 0.0, 0.0],
            [0.0, 1 / 3, 1 / 3, 2 / 3],
            [0.2, 0.4, 0.5, 0.8],
        ),
        (  # mixed truth types
            [1, False, 1, True],
            [0.5, 0.4, 0.8, 0.2],
            [1.0, 1.0, 0.0, 0.0],
            [0.0, 1 / 3, 1 / 3, 2 / 3],
            [0.2, 0.4, 0.5, 0.8],
        ),
        # Float values not allowed in truth
        pytest.param(
            [1, 0.5, 0, 0],
            [1, 1, 0, 0],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_detection_error_tradeoff(
    truth, prediction, expected_fmr, expected_fnmr, expected_thresholds
):
    ret = audmetric.detection_error_tradeoff(truth, prediction)
    # Check return types
    assert len(ret) == 3  # fmr, fnmr, thresholds
    for arr in ret:
        assert type(arr) == np.ndarray
        for val in arr:
            assert isinstance(val, np.floating)
    # Check return values
    fmr, fnmr, thresholds = ret
    np.testing.assert_almost_equal(fmr, expected_fmr)
    np.testing.assert_almost_equal(fnmr, expected_fnmr)
    np.testing.assert_almost_equal(thresholds, expected_thresholds)


@pytest.mark.parametrize(
    "truth, prediction, expected_eer, expected_threshold",
    [
        (
            [1, 1, 0, 0],
            [1, 1, 0, 0],
            0,
            1,
        ),
        (
            [True, True, False, False],
            [1, 1, 0, 0],
            0,
            1,
        ),
        (
            [True, True, False, False],
            [True, True, False, False],
            0,
            1,
        ),
        (
            [1, 1, 0, 0],
            [0.9, 0.9, 0.1, 0.1],
            0,
            0.9,
        ),
        (
            [1, 1, 0, 0],
            [1, 0.1, 0.1, 0],
            0.25,
            0.1,
        ),
        (
            [1, 1, 0, 0],
            [0.8, 0.7, 0.4, 0.1],
            0,
            0.7,
        ),
        (
            [1, 1, 0, 0, 0],
            [0.8, 0.7, 0.4, 0.1, 0.1],
            0,
            0.7,
        ),
        # Non integer truth not allowed
        pytest.param(
            [0.9, 0.9, 0.1, 0.1],
            [0.9, 0.9, 0.1, 0.1],
            0,
            1,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_equal_error_rate(truth, prediction, expected_eer, expected_threshold):
    r"""Test audmetric.equal_error_rate().

    The expected values for EER and the thresholds
    were calculated by the ``pyeer`` package,
    see https://github.com/manuelaguadomtz/pyeer.

    """
    eer, stats = audmetric.equal_error_rate(truth, prediction)
    # Check expected results
    assert isinstance(eer, float)
    assert isinstance(stats.threshold, float)
    assert eer == expected_eer
    assert stats.threshold == expected_threshold


def test_equal_error_rate_warnings():
    # No imposter scores (division by 0)
    truth = np.array([1, 1])
    prediction = np.array([1, 1])
    warning = "invalid value encountered"
    with pytest.warns(RuntimeWarning, match=warning):
        eer, stats = audmetric.equal_error_rate(truth, prediction)

    # Curves to not overlap
    truth = np.array([1, 1, 0])
    prediction = np.array([0.5, 0.5, 0.5])
    warning = (
        r"false match rate and false non-match rate curves "
        r"do not intersect each other"
    )
    with pytest.warns(RuntimeWarning, match=warning):
        eer, stats = audmetric.equal_error_rate(truth, prediction)


@pytest.mark.parametrize(
    "truth,prediction,eer",
    [
        ([], [], 0),
        ([[]], [[]], 0),
        ([[None]], [[]], 1.0),
        ([[None]], [[1]], 1.0),
        ([[None]], [[1, 2]], 1.0),
        ([[0], []], [[1], []], 0.5),
        ([[0, 1]], [[0]], 0.5),
        ([[0]], [[0, 1]], 0.5),
        ([[0, 1], [2]], [[0], [2]], 0.25),
        pytest.param(
            [[0, 1]], [[0], [2]], 0.0, marks=pytest.mark.xfail(raises=ValueError)
        ),
        ("lorem", "lorm", 0.2),
        (["lorem"], ["lorm"], 0.2),
        (["lorem", "ipsum"], ["lorm", "ipsum"], 0.1),
        pytest.param(
            ["lorem", "ipsum"],
            ["lorm"],
            0.0,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_event_error_rate(truth, prediction, eer):
    np.testing.assert_equal(audmetric.event_error_rate(truth, prediction), eer)


@pytest.mark.parametrize(
    "class_range,num_elements,to_string,percentage",
    [
        ([0, 10], 5, False, False),
        ([0, 10], 1, False, False),
        ([0, 10], 10, False, False),
        ([0, 2], 100, False, False),
        ([0, 10], 10, True, False),
        ([0, 10], 100, True, True),
        ([0, 10], 10, True, True),
    ],
)
def test_confusion_matrix(class_range, num_elements, to_string, percentage):
    t = np.random.randint(class_range[0], class_range[1], size=num_elements)
    p = np.random.randint(class_range[0], class_range[1], size=num_elements)

    if to_string:
        t = [str(w) for w in t]
        p = [str(w) for w in p]

    cm = audmetric.confusion_matrix(t, p, normalize=percentage)

    if percentage:
        cm_sklearn = sklearn.metrics.confusion_matrix(t, p, normalize="true")
    else:
        cm_sklearn = sklearn.metrics.confusion_matrix(t, p)

    np.testing.assert_almost_equal(cm, cm_sklearn)


@pytest.mark.parametrize(
    "truth,prediction,edit_distance",
    [
        ("lorem", "lorem", 0),
        ("lorem", "", 5),
        ("", "lorem", 5),
        ("lorem", "lorm", 1),
        ("lorem", "lorrem", 1),
        ("lorem", "lorom", 1),
        ("lorem", "morel", 2),
        ([], [0], 1),
        ([0], [], 1),
        ([0, 1, 2], [0, 1], 1),
        ([0, 1, 2], [0, 1, 1], 1),
        ([None], [], 1),
        ([None], [1], 1),
        ([None], [1, 2], 2),
    ],
)
def test_edit_distance(truth, prediction, edit_distance):
    np.testing.assert_equal(audmetric.edit_distance(truth, prediction), edit_distance)


@pytest.mark.parametrize(
    "value_range,num_elements",
    [
        ([0, 10], 5),
        ([0, 10], 1),
        ([0, 10], 10),
        ([0, 2], 100),
    ],
)
def test_mean_absolute_error(value_range, num_elements):
    t = np.random.randint(value_range[0], value_range[1], size=num_elements)
    p = np.random.randint(value_range[0], value_range[1], size=num_elements)

    np.testing.assert_almost_equal(
        audmetric.mean_absolute_error(t, p),
        sklearn.metrics.mean_absolute_error(t, p),
    )

    t = pd.Series(t).astype("Int64")
    t = pd.Series(t).astype("Int64")

    np.testing.assert_almost_equal(
        audmetric.mean_absolute_error(t, p),
        sklearn.metrics.mean_absolute_error(t, p),
    )


@pytest.mark.parametrize(
    "value_range,num_elements",
    [
        ([0, 10], 5),
        ([0, 10], 1),
        ([0, 10], 10),
        ([0, 2], 100),
    ],
)
def test_mean_squared_error(value_range, num_elements):
    t = np.random.randint(value_range[0], value_range[1], size=num_elements)
    p = np.random.randint(value_range[0], value_range[1], size=num_elements)

    np.testing.assert_almost_equal(
        audmetric.mean_squared_error(t, p),
        sklearn.metrics.mean_squared_error(t, p),
    )

    t = pd.Series(t).astype("Int64")
    t = pd.Series(t).astype("Int64")

    np.testing.assert_almost_equal(
        audmetric.mean_absolute_error(t, p),
        sklearn.metrics.mean_absolute_error(t, p),
    )


@pytest.mark.parametrize(
    "truth,prediction",
    [
        (
            np.random.randint(0, 10, size=5),
            np.random.randint(0, 10, size=5),
        ),
        (
            pd.Series(np.random.randint(0, 10, size=5)).astype("Int64"),
            pd.Series(np.random.randint(0, 10, size=5)).astype("Int64"),
        ),
        (
            np.random.randint(0, 10, size=1),
            np.random.randint(0, 10, size=1),
        ),
        (
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=10),
        ),
        (
            np.random.randint(0, 2, size=100),
            np.random.randint(0, 2, size=100),
        ),
        (
            np.array([]),
            np.array([]),
        ),
        (
            np.zeros(10),
            np.zeros(10),
        ),
    ],
)
def test_pearson_cc(truth, prediction):
    if len(prediction) < 2 or prediction.std() == 0:
        pcc = np.nan
    else:
        pcc = np.corrcoef(list(truth), list(prediction))[0][1]
    np.testing.assert_almost_equal(
        audmetric.pearson_cc(truth, prediction),
        pcc,
    )


@pytest.mark.parametrize(
    "truth, prediction, labels, zero_division",
    [
        (
            ["a"],
            ["a"],
            None,
            0,
        ),
        (
            ["a"],
            ["b"],
            None,
            0,
        ),
        (
            ["a"],
            ["b"],
            ["a", "b"],
            0,
        ),
        (
            ["a"],
            ["b"],
            ["a", "b"],
            1,
        ),
        (
            ["a", "b"],
            ["b", "a"],
            None,
            0,
        ),
        (
            np.random.randint(0, 10, 5),
            np.random.randint(0, 10, 5),
            None,
            0,
        ),
        (
            pd.Series(np.random.randint(0, 10, 5)).astype("Int64"),
            pd.Series(np.random.randint(0, 10, 5)).astype("Int64"),
            None,
            0,
        ),
        (
            np.random.randint(0, 10, 100),
            np.random.randint(0, 10, 100),
            None,
            0,
        ),
    ],
)
def test_recall_precision_fscore(truth, prediction, labels, zero_division):
    for metric, sklearn_metric in (
        (
            audmetric.unweighted_average_recall,
            sklearn.metrics.recall_score,
        ),
        (
            audmetric.unweighted_average_precision,
            sklearn.metrics.precision_score,
        ),
        (
            audmetric.unweighted_average_fscore,
            sklearn.metrics.f1_score,
        ),
    ):
        result = metric(
            truth,
            prediction,
            labels,
            zero_division=zero_division,
        )
        expected = (
            sklearn_metric(
                list(truth),
                list(prediction),
                average="macro",
                zero_division=zero_division,
            ),
        )
        np.testing.assert_almost_equal(
            result,
            expected,
        )


@pytest.mark.parametrize(
    "truth,prediction,protected_variable,metric,labels,subgroups,"
    "zero_division,expected",
    [
        pytest.param(
            [],
            [],
            [],
            None,
            [],
            [0],
            None,
            {},
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        ([], [], [], audmetric.recall_per_class, [], [], 0.0, {}),
        (
            [1],
            [0],
            [0],
            audmetric.recall_per_class,
            [0, 1],
            [0],
            0.0,
            {0: {0: 0.0, 1: 0.0}},
        ),
        (
            [1],
            [0],
            [0],
            audmetric.precision_per_class,
            [0, 1],
            [0],
            0.0,
            {0: {0: 0.0, 1: 0.0}},
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            audmetric.recall_per_class,
            [0, 1],
            [0, 1],
            np.nan,
            {0: {0: np.nan, 1: 0.0}, 1: {0: np.nan, 1: 1.0}},
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            audmetric.recall_per_class,
            [1],
            [0, 1],
            np.nan,
            {0: {1: np.nan}, 1: {1: 1.0}},
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            audmetric.precision_per_class,
            [0, 1],
            [0, 1],
            np.nan,
            {0: {0: 0.0, 1: np.nan}, 1: {0: np.nan, 1: 1.0}},
        ),
    ],
)
def test_scores_per_subgroup_and_class(
    truth,
    prediction,
    protected_variable,
    metric,
    labels,
    subgroups,
    zero_division,
    expected,
):
    np.testing.assert_equal(
        audmetric.core.utils.scores_per_subgroup_and_class(
            truth,
            prediction,
            protected_variable,
            metric,
            labels=labels,
            subgroups=subgroups,
            zero_division=zero_division,
        ),
        expected,
    )


@pytest.mark.parametrize(
    "truth,prediction,protected_variable,labels,subgroups,metric,reduction," "expected",
    [
        (
            [],
            [],
            [],
            None,
            None,
            audmetric.fscore_per_class,
            lambda x: abs(x[0] - x[1]),
            np.nan,
        ),
        (
            [0],
            [0],
            [0],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            np.nan,
        ),
        pytest.param(
            [0, 0],
            [0],
            [0],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            [0, 0],
            [0, 0],
            [0, 0],
            None,
            [0, 1],
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        (
            [0, 0],
            [0, 0],
            [0, 1],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            0.0,
        ),
        (
            [0, 1],
            [0, 0],
            [0, 1],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            np.nan,
        ),
        (
            [0, 1],
            [0, 0],
            [0, 1],
            None,
            None,
            audmetric.precision_per_class,
            lambda x: abs(x[0] - x[1]),
            1.0,
        ),
        (
            [0, 1],
            [0, 0],
            [0, 1],
            [0, 1, 2, 3],
            None,
            audmetric.precision_per_class,
            lambda x: abs(x[0] - x[1]),
            1.0,
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            1.0,
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: x[0] - x[1],
            -1.0,
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            None,
            [1, 0],
            audmetric.recall_per_class,
            lambda x: x[0] - x[1],
            1.0,
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            None,
            [1, 0],
            audmetric.fscore_per_class,
            lambda x: abs(x[0] - x[1]),
            1.0,
        ),
        (
            [1, 1],
            [0, 1],
            [0, 1],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            1.0,
        ),
        (
            [1, 1, 2],
            [0, 1, 2],
            [0, 1, 1],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            1.0,
        ),
        (
            [1, 1, 2, 2],
            [0, 1, 2, 2],
            [0, 1, 1, 0],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: abs(x[0] - x[1]),
            0.5,
        ),
        (
            [1, 1, 2, 2],
            [0, 1, 2, 2],
            [0, 1, 1, 0],
            None,
            None,
            audmetric.recall_per_class,
            lambda x: x[0] - x[1],
            -0.5,
        ),
        (
            [0, 0, 0, 0],
            [1, 1, 0, 0],
            [0, 1, 2, 3],
            None,
            None,
            audmetric.recall_per_class,
            np.std,
            0.5,
        ),
    ],
)
def test_unweighted_average_bias(
    truth,
    prediction,
    protected_variable,
    labels,
    subgroups,
    metric,
    reduction,
    expected,
):
    np.testing.assert_equal(
        audmetric.unweighted_average_bias(
            truth,
            prediction,
            protected_variable,
            labels=labels,
            subgroups=subgroups,
            metric=metric,
            reduction=reduction,
        ),
        expected,
    )


def test_recall_precision_fscore_nan():
    truth = ["a", "b"]
    prediction = ["a", "a"]
    labels = ["a", "b", "c"]

    for metric in (
        audmetric.unweighted_average_recall,
        audmetric.unweighted_average_precision,
        audmetric.unweighted_average_fscore,
    ):
        result = metric(
            truth,
            prediction,
            labels,
            zero_division=np.nan,
        )
        np.testing.assert_equal(result, np.nan)

    result = audmetric.recall_per_class(
        truth,
        prediction,
        labels,
        zero_division=np.nan,
    )
    np.testing.assert_equal(result, {"a": 1.0, "b": 0.0, "c": np.nan})

    result = audmetric.precision_per_class(
        truth,
        prediction,
        labels,
        zero_division=np.nan,
    )
    np.testing.assert_equal(result, {"a": 0.5, "b": np.nan, "c": np.nan})

    result = audmetric.fscore_per_class(
        truth,
        prediction,
        labels,
        zero_division=np.nan,
    )
    np.testing.assert_equal(result, {"a": 2 / 3, "b": 0.0, "c": np.nan})


@pytest.mark.parametrize(
    "truth, prediction, labels, zero_division, expected",
    [
        (
            ["a"],
            ["a"],
            ["a", "b"],
            np.nan,
            np.nan,
        ),
        (
            ["a"],
            ["a"],
            ["a", "b"],
            0,
            0.5,
        ),
        (
            ["a"],
            ["a"],
            ["a", "b"],
            1,
            1,
        ),
    ],
)
def test_unweighted_average_zero_division(
    truth, prediction, labels, zero_division, expected
):
    for metric in (
        audmetric.unweighted_average_recall,
        audmetric.unweighted_average_precision,
        audmetric.unweighted_average_fscore,
    ):
        result = metric(
            truth,
            prediction,
            labels,
            zero_division=zero_division,
        )
        np.testing.assert_equal(result, expected)


@pytest.mark.parametrize(
    "weights,num_elements,to_string",
    [
        (
            [
                [0, 1, 2],
                [1, 0, 1],
                [2, 1, 0],
            ],
            5,
            False,
        ),
        (
            [
                [0, 1, 2],
                [1, 0, 1],
                [2, 1, 0],
            ],
            5,
            True,
        ),
        pytest.param(  # shape of cm and weights do not match
            [
                [0, 1, 2],
                [1, 0, 1],
            ],
            5,
            False,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_weighted_confusion_error(weights, num_elements, to_string):
    n = len(weights)
    t = np.zeros(num_elements, dtype=int)
    t[:n] = range(n)
    p = np.random.randint(0, len(weights), size=num_elements)

    if to_string:
        t = [str(w) for w in t]
        p = [str(w) for w in p]

    wce = audmetric.weighted_confusion_error(t, p, weights)
    cm = audmetric.confusion_matrix(t, p, normalize=True)
    weights = weights / np.sum(weights)
    np.testing.assert_equal(
        wce,
        float(np.sum(cm * weights)),
    )


@pytest.mark.parametrize(
    "truth,prediction,wer",
    [
        ([[]], [[]], 0),
        ([[None]], [[]], 1.0),
        ([[None]], [["lorem"]], 1.0),
        ([[None]], [["lorem", "ipsum"]], 1.0),
        ([["lorem"]], [[]], 1),
        ([[]], [["lorem"]], 1),
        ([["lorem", "ipsum"]], [["lorem"]], 0.5),
        ([["lorem"]], [["lorem", "ipsum"]], 0.5),
        ([["lorem"]], [["lorem"]], 0),
        ([["lorem", "ipsum"]], [["lorm", "ipsum"]], 0.5),
        (
            [["lorem", "ipsum"], ["north", "wind", "and", "sun"]],
            [["lorm", "ipsum"], ["north", "wind"]],
            0.5,
        ),
        pytest.param(
            [["lorem"], []],
            [[]],
            0.0,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_word_error_rate(truth, prediction, wer):
    np.testing.assert_equal(audmetric.word_error_rate(truth, prediction), wer)



import numpy as np
import pandas as pd
import pytest

import audmetric


def expected_ccc(truth, prediction):
    r"""Expected Concordance Correlation Coefficient.

    This is a direct implementation of its math equation.

    """
    prediction = np.array(list(prediction))
    truth = np.array(list(truth))

    denominator = (
        prediction.std() ** 2
        + truth.std() ** 2
        + (prediction.mean() - truth.mean()) ** 2
    )
    if denominator == 0:
        ccc = np.nan
    else:
        r = np.corrcoef(list(prediction), list(truth))[0][1]
        numerator = 2 * r * prediction.std() * truth.std()
        ccc = numerator / denominator

    return ccc


@pytest.mark.parametrize("ignore_nan", [True, False])
@pytest.mark.parametrize(
    "truth, prediction",
    [
        # NOTE: this test assumes
        # that all truth and prediction values
        # do not contain NaN
        (
            np.random.randint(0, 10, size=5),
            np.random.randint(0, 10, size=5),
        ),
        (
            pd.Series(np.random.randint(0, 10, size=5)).astype("Int64"),
            pd.Series(np.random.randint(0, 10, size=5)).astype("Int64"),
        ),
        (
            np.random.randint(0, 10, size=1),
            np.random.randint(0, 10, size=1),
        ),
        (
            np.random.randint(0, 10, size=10),
            np.random.randint(0, 10, size=10),
        ),
        (
            np.random.randint(0, 2, size=100),
            np.random.randint(0, 2, size=100),
        ),
        (
            np.array([]),
            np.array([]),
        ),
        (
            np.zeros(10),
            np.zeros(10),
        ),
    ],
)
def test_concordance_cc(truth, prediction, ignore_nan):
    ccc = audmetric.concordance_cc(truth, prediction, ignore_nan=ignore_nan)

    np.testing.assert_almost_equal(
        ccc,
        expected_ccc(truth, prediction),
    )


@pytest.mark.parametrize(
    "truth, prediction, ignore_nan, expected",
    [
        (
            [],
            [],
            True,
            np.nan,
        ),
        (
            [],
            [],
            False,
            np.nan,
        ),
        (
            [0],
            [0],
            True,
            np.nan,
        ),
        (
            [0],
            [0],
            False,
            np.nan,
        ),
        (
            [0, np.nan],
            [0, np.nan],
            True,
            np.nan,
        ),
        (
            [0, np.nan],
            [0, np.nan],
            False,
            np.nan,
        ),
        (
            [0, 1, 2, 3],
            [1, 2, 3, 4],
            True,
            expected_ccc([0, 1, 2, 3], [1, 2, 3, 4]),
        ),
        (
            [np.nan, 1, 2, 3],
            [np.nan, 2, 3, 4],
            True,
            expected_ccc([1, 2, 3], [2, 3, 4]),
        ),
        (
            [np.nan, 1, 2, 3],
            [1, 2, 3, np.nan],
            True,
            expected_ccc([1, 2], [2, 3]),
        ),
        (
            [0, np.nan, 2, 3],
            [1, 2, 3, 4],
            True,
            expected_ccc([0, 2, 3], [1, 3, 4]),
        ),
        (
            [0, 1, 2, 3],
            [1, 2, np.nan, 4],
            True,
            expected_ccc([0, 1, 3], [1, 2, 4]),
        ),
        (
            [np.nan, np.nan, 2, 3],
            [1, 2, 3, np.nan],
            True,
            expected_ccc([2], [3]),
        ),
        (
            [np.nan, np.nan, 2, 3],
            [1, 2, 3, np.nan],
            False,
            np.nan,
        ),
    ],
)
def test_concordance_cc_ignore_nan(
    truth,
    prediction,
    ignore_nan,
    expected,
):
    ccc = audmetric.concordance_cc(truth, prediction, ignore_nan=ignore_nan)

    np.testing.assert_almost_equal(
        ccc,
        expected,
    )



from audmetric import utils
from audmetric.core.api import accuracy
from audmetric.core.api import concordance_cc
from audmetric.core.api import confusion_matrix
from audmetric.core.api import detection_error_tradeoff
from audmetric.core.api import edit_distance
from audmetric.core.api import equal_error_rate
from audmetric.core.api import event_error_rate
from audmetric.core.api import fscore_per_class
from audmetric.core.api import linkability
from audmetric.core.api import mean_absolute_error
from audmetric.core.api import mean_squared_error
from audmetric.core.api import pearson_cc
from audmetric.core.api import precision_per_class
from audmetric.core.api import recall_per_class
from audmetric.core.api import unweighted_average_bias
from audmetric.core.api import unweighted_average_fscore
from audmetric.core.api import unweighted_average_precision
from audmetric.core.api import unweighted_average_recall
from audmetric.core.api import weighted_confusion_error
from audmetric.core.api import word_error_rate


__all__ = []


# Dynamically get the version of the installed module
try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__name__)
except Exception:  # pragma: no cover
    importlib = None  # pragma: no cover
finally:
    del importlib



import numpy as np
import pytest

import audeer

import audmetric


# The expected values were manually calculated using code from
# https://gitlab.inria.fr/magnet/anonymization_metrics.
# See tests/assests/linkability/README.md
# for instructions,
# and results in tests/assests/linkability/results.csv
#
# To ensure reproducibility of the signals
# we create our own signal generator functions
#
def truth(samples=1000):
    r"""Generate truth values for linkability test.

    The truth values are in the form:
    ``[0, 1, 0, 1, ..., 0, 1]``

    """
    return [0, 1] * samples


def prediction(range1, range2, samples=1000, random_state=1):
    r"""Generate prediction values for linkability test.

    The values are returned as pairs
    drawn from uniform distribution
    given by ``range1`` and ``range2``, e.g.
    ``[0.12, 0.84, 0.07, 0.81, ..., 0.13, 0.84]``

    """
    np.random.seed(random_state)
    return audeer.flatten_list(
        [
            [np.random.uniform(*range1), np.random.uniform(*range2)]
            for _ in range(samples)
        ]
    )


@pytest.mark.parametrize(
    "truth, prediction, omega, expected",
    [
        # All following tests get [0, 1, 0, 1, ..., 0, 1] as truth values.
        # The corresponding prediction values
        # are drawn from two uniform distributions,
        # from Distribution 1 for all non-mated entries
        # (matching 0 positions from truth)
        # and from Distribution 2 for all mated entries
        # (matching 1 positions from truth).
        # The ranges of the two distributions
        # are provided as arguments to the `prediction()` function
        # and indicated by the ____##____ distribution sketches.
        # The guessing probability for identifying the mated speakers
        # is given for each test,
        # and determines the `omega` value.
        # The expected values were pre-calculated by
        # `tests/assests/linkability/linkability_reference.py`.
        (
            # Distribution 1: ##________
            # Distribution 2: ________##
            # Guessing: 0.5
            truth(),
            prediction((0.0, 0.2), (0.8, 1.0)),
            1,
            0.9710,
        ),
        (
            # Distribution 1: ##________
            # Distribution 2: ________##
            # Guessing: 0.1
            truth(),
            prediction((0.0, 0.2), (0.8, 1.0)),
            1 / 9.0,
            0.9710,
        ),
        (
            # Distribution 1: _##_______
            # Distribution 2: _______##_
            # Guessing: 0.5
            truth(),
            prediction((0.1, 0.3), (0.7, 0.9)),
            1,
            0.9775,
        ),
        (
            # Distribution 1: __##______
            # Distribution 2: ______##__
            # Guessing: 0.5
            truth(),
            prediction((0.2, 0.4), (0.6, 0.8)),
            1,
            0.9840,
        ),
        (
            # Distribution 1: ___##_____
            # Distribution 2: _____##___
            # Guessing: 0.5
            truth(),
            prediction((0.3, 0.5), (0.5, 0.7)),
            1,
            0.9905,
        ),
        (
            # Distribution 1: ____##____
            # Distribution 2: _____##___
            # Guessing: 0.5
            truth(),
            prediction((0.4, 0.6), (0.5, 0.7)),
            1,
            0.5619,
        ),
        (
            # Distribution 1: ___##_____
            # Distribution 2: ____##____
            # Guessing: 0.5
            truth(),
            prediction((0.3, 0.5), (0.4, 0.6)),
            1,
            0.5619,
        ),
        (
            # Distribution 1: ____##____
            # Distribution 2: ____##____
            # Guessing: 0.5
            truth(),
            prediction((0.4, 0.6), (0.4, 0.6)),
            1,
            0.1352,
        ),
        (
            # Distribution 1: _____##___
            # Distribution 2: ___##_____
            # Guessing: 0.5
            truth(),
            prediction((0.5, 0.7), (0.3, 0.5)),
            1,
            0.9870,
        ),
        (
            # Distribution 1: ___###____
            # Distribution 2: _____###__
            # Guessing: 0.5
            truth(),
            prediction((0.3, 0.6), (0.5, 0.8)),
            1,
            0.7066,
        ),
        (
            # Distribution 1: __####____
            # Distribution 2: _____####_
            # Guessing: 0.5
            truth(),
            prediction((0.2, 0.6), (0.5, 0.9)),
            1,
            0.7654,
        ),
        (
            # Distribution 1: _#####____
            # Distribution 2: _____#####
            # Guessing: 0.5
            truth(),
            prediction((0.1, 0.6), (0.5, 1.0)),
            1,
            0.8083,
        ),
        (
            # Distribution 1: _#####____
            # Distribution 2: _____#####
            # Guessing: 0.5
            truth(),
            prediction((0.4, 0.6), (0.5, 0.7)),
            1,
            0.5619,
        ),
        (
            # Distribution 1: ###########
            # Distribution 2: ###########
            # Guessing: 0.5
            truth(),
            prediction((0.0, 1.0), (0.0, 1.0)),
            1.0,
            0.1352,
        ),
        (
            # Distribution 1: ###########
            # Distribution 2: ###########
            # Guessing: 0.33
            truth(),
            prediction((0.0, 1.0), (0.0, 1.0)),
            1 / 2.0,
            0.0167,
        ),
        (
            # Distribution 1: ###########
            # Distribution 2: ###########
            # Guessing: 0.66
            truth(),
            prediction((0.0, 1.0), (0.0, 1.0)),
            2.0 / 1,
            0.3746,
        ),
    ],
)
def test_linkability(truth, prediction, omega, expected):
    linkability = audmetric.linkability(truth, prediction, omega=omega)
    np.testing.assert_allclose(linkability, expected, rtol=0.001)



import collections
import typing
import warnings

import numpy as np

import audeer

from audmetric.core.utils import assert_equal_length
from audmetric.core.utils import infer_labels
from audmetric.core.utils import scores_per_subgroup_and_class


def accuracy(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Union[str, int]] = None,
) -> float:
    r"""Classification accuracy.

    .. math::

        \text{accuracy} = \frac{\text{number of correct predictions}}
                               {\text{number of total predictions}}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            Sample is considered in computation if either prediction or
            ground truth (logical OR) is contained in labels.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.

    Returns:
        accuracy of prediction :math:`\in [0, 1]`

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> accuracy([0, 0], [0, 1])
        0.5

    """
    assert_equal_length(truth, prediction)
    if labels is None:
        labels = infer_labels(truth, prediction)

    prediction = np.array(prediction)
    truth = np.array(truth)

    # keep where both prediction and truth contained in `labels`
    label_mask = np.nonzero(
        np.logical_or(np.isin(truth, labels), np.isin(prediction, labels))
    )
    truth = truth[label_mask]
    prediction = prediction[label_mask]

    if len(prediction) == 0:
        return np.nan
    else:
        return float(sum(prediction == truth) / len(prediction))


def concordance_cc(
    truth: typing.Sequence[float],
    prediction: typing.Sequence[float],
    *,
    ignore_nan: bool = False,
) -> float:
    r"""Concordance correlation coefficient.

    .. math::

        \rho_c = \frac{2\rho\sigma_\text{prediction}\sigma_\text{truth}}
                      {\sigma_\text{prediction}^2 + \sigma_\text{truth}^2 + (
                      \mu_\text{prediction}-\mu_\text{truth})^2}

    where :math:`\rho` is the Pearson correlation coefficient,
    :math:`\mu` the mean
    and :math:`\sigma^2` the variance.\ :footcite:`Lin1989`

    .. footbibliography::

    Args:
        truth: ground truth values
        prediction: predicted values
        ignore_nan: if ``True``
            all samples that contain ``NaN``
            in ``truth`` or ``prediction``
            are ignored

    Returns:
        concordance correlation coefficient :math:`\in [-1, 1]`

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> concordance_cc([0, 1, 2], [0, 1, 1])
        0.6666666666666665

    """
    assert_equal_length(truth, prediction)

    if not isinstance(truth, np.ndarray):
        truth = np.array(list(truth))
    if not isinstance(prediction, np.ndarray):
        prediction = np.array(list(prediction))

    if ignore_nan:
        mask = ~(np.isnan(truth) | np.isnan(prediction))
        truth = truth[mask]
        prediction = prediction[mask]

    if len(prediction) < 2:
        return np.nan

    length = prediction.size
    mean_y = np.mean(truth)
    mean_x = np.mean(prediction)
    a = prediction - mean_x
    b = truth - mean_y

    numerator = 2 * np.dot(a, b)
    denominator = np.dot(a, a) + np.dot(b, b) + length * (mean_x - mean_y) ** 2

    if denominator == 0:
        ccc = np.nan
    else:
        ccc = numerator / denominator

    return float(ccc)


def confusion_matrix(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    normalize: bool = False,
) -> typing.List[typing.List[typing.Union[int, float]]]:
    r"""Confusion matrix.

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        normalize: normalize confusion matrix over the rows

    Returns:
        confusion matrix

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> truth = [0, 1, 2]
        >>> prediction = [0, 2, 0]
        >>> confusion_matrix(truth, prediction)
        [[1, 0, 0], [0, 0, 1], [1, 0, 0]]

    """
    assert_equal_length(truth, prediction)
    if labels is None:
        labels = infer_labels(truth, prediction)

    truth = np.array(truth)
    prediction = np.array(prediction)

    matrix = []
    for row in labels:
        row_indices = np.where(truth == row)
        y_row = prediction[row_indices]
        row_matrix = []
        for column in labels:
            row_matrix += [len(np.where(y_row == column)[0])]
        matrix += [row_matrix]

    if normalize:
        for idx, row in enumerate(matrix):
            if np.sum(row) != 0:
                row_sum = float(np.sum(row))
                matrix[idx] = [x / row_sum for x in row]

    return matrix


def detection_error_tradeoff(
    truth: typing.Sequence[typing.Union[bool, int]],
    prediction: typing.Sequence[typing.Union[bool, int, float]],
) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""Detection error tradeoff for verification experiments.

    The `detection error tradeoff (DET)`_
    is a graph showing
    the false non-match rate (FNMR)
    against the false match rate (FMR).
    The FNMR indicates
    how often an enrolled speaker was missed.
    The FMR indicates
    how often an impostor was verified as the enrolled speaker.

    This function does not return a figure,
    but the FMR and FNMR,
    together with the corresponding verification thresholds
    at which a similarity value
    was regarded to belong to the enrolled speaker.

    ``truth`` may only contain entries like ``[1, 0, True, False...]``,
    whereas prediction values
    can also contain similarity scores, e.g. ``[0.8, 0.1, ...]``.

    The implementation was inspired by pyeer.eer_stats.calculate_roc but has
    been accelerated by using numpy-arrays instead of lists.

    .. _detection error tradeoff (DET): https://en.wikipedia.org/wiki/Detection_error_tradeoff
    .. _pyeer: https://github.com/manuelaguadomtz/pyeer

    Args:
        truth: ground truth classes
        prediction: predicted classes or similarity scores

    Returns:
        * false match rate (FMR)
        * false non-match rate (FNMR)
        * verification thresholds

    Raises:
        ValueError: if ``truth`` contains values
            different from ``1, 0, True, False``

    Examples:
        >>> truth = [1, 0]
        >>> prediction = [0.9, 0.1]
        >>> detection_error_tradeoff(truth, prediction)
        (array([1., 0.]), array([0., 0.]), array([0.1, 0.9]))

    """  # noqa: E501
    # Get mated scores
    # (genuine matching scores)
    # and non-mated scores
    # (impostor matching scores)
    gscores, iscores = _matching_scores(truth, prediction)

    gscores_number = len(gscores)
    iscores_number = len(iscores)

    # Labeling genuine scores as 1 and impostor scores as 0
    gscores = np.column_stack((gscores, np.ones(gscores_number, dtype=int)))
    iscores = np.column_stack((iscores, np.zeros(iscores_number, dtype=int)))

    # Stacking scores
    all_scores = np.concatenate([gscores, iscores])
    sorted_indices = np.argsort(all_scores[:, 0])
    scores = all_scores[sorted_indices]
    cumul = np.cumsum(scores[:, 1])

    # Grouping scores
    thresholds, u_indices = np.unique(scores[:, 0], return_index=True)

    # Calculating FNM and FM distributions
    fnm = cumul[u_indices] - scores[u_indices][:, 1]  # rejecting s < t
    fm = iscores_number - (u_indices - fnm)

    # Calculating FMR and FNMR
    fnmr = fnm / gscores_number
    fmr = fm / iscores_number

    return fmr, fnmr, thresholds


def edit_distance(
    truth: typing.Union[str, typing.Sequence[int]],
    prediction: typing.Union[str, typing.Sequence[int]],
) -> int:
    r"""Edit distance between two sequences of chars or ints.

    The implementation follows the `Wagner-Fischer algorithm`_.

    .. _Wagner-Fischer algorithm:
        https://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm

    Args:
        truth: ground truth sequence
        prediction: predicted sequence

    Returns:
        edit distance

    Examples:
        >>> truth = "lorem"
        >>> prediction = "lorm"
        >>> edit_distance(truth, prediction)
        1
        >>> truth = [0, 1, 2]
        >>> prediction = [0, 1]
        >>> edit_distance(truth, prediction)
        1

    """
    if truth == prediction:
        return 0

    elif len(prediction) == 0:
        return len(truth)

    elif len(truth) == 0:
        return len(prediction)

    m0 = [None] * (len(truth) + 1)
    m1 = [None] * (len(truth) + 1)

    for i in range(len(m0)):
        m0[i] = i

    for i in range(len(prediction)):
        m1[0] = i + 1
        for j in range(len(truth)):
            cost = 0 if prediction[i] == truth[j] else 1
            m1[j + 1] = min(
                m1[j] + 1,  # deletion
                m0[j + 1] + 1,  # insertion
                m0[j] + cost,
            )  # substitution

        for j in range(len(m0)):
            m0[j] = m1[j]

    return m1[len(truth)]


def equal_error_rate(
    truth: typing.Sequence[typing.Union[bool, int]],
    prediction: typing.Sequence[typing.Union[bool, int, float]],
) -> typing.Tuple[float, collections.namedtuple]:
    r"""Equal error rate for verification tasks.

    The equal error rate (EER) is the point
    where false non-match rate (FNMR)
    and the impostors or false match rate (FMR)
    are identical.
    The FNMR indicates
    how often an enrolled speaker was missed.
    The FMR indicates
    how often an impostor was verified as the enrolled speaker.

    In practice the score distribution is not continuous
    and an interval is returned instead.
    The EER value will be set as the midpoint
    of this interval::footcite:`Maio2002`

    .. math::

        \text{EER} = \frac{
            \min(\text{FNMR}[t], \text{FMR}[t])
            + \max(\text{FNMR}[t], \text{FMR}[t])
        }{2}

    with :math:`t = \text{argmin}(|\text{FNMR} - \text{FMR}|)`.

    ``truth`` may only contain entries like ``[1, 0, True, False...]``,
    whereas prediction values
    can also contain similarity scores, e.g. ``[0.8, 0.1, ...]``.

    The implementation is identical with the one provided
    by the pyeer_ package.

    .. footbibliography::

    .. _pyeer: https://github.com/manuelaguadomtz/pyeer

    Args:
        truth: ground truth classes
        prediction: predicted classes or similarity scores

    Returns:
        * equal error rate (EER)
        * namedtuple containing
          ``fmr``,
          ``fnmr``,
          ``thresholds``,
          ``threshold``
          whereas the last one corresponds to the threshold
          corresponding to the returned EER

    Raises:
        ValueError: if ``truth`` contains values
            different from ``1, 0, True, False``

    Examples:
        >>> truth = [0, 1, 0, 1, 0]
        >>> prediction = [0.2, 0.8, 0.4, 0.5, 0.5]
        >>> eer, stats = equal_error_rate(truth, prediction)
        >>> eer
        0.16666666666666666
        >>> stats.threshold
        0.5

    """
    Stats = collections.namedtuple(
        "stats",
        [
            "fmr",  # False match rates (FMR)
            "fnmr",  # False non-match rates (FNMR)
            "thresholds",  # Thresholds
            "threshold",  # verification threshold for EER
        ],
    )
    fmr, fnmr, thresholds = detection_error_tradeoff(truth, prediction)
    diff = fmr - fnmr
    # t1 and t2 are our time indices
    t2 = np.where(diff <= 0)[0]
    if len(t2) > 0:
        t2 = t2[0]
    else:
        warnings.warn(
            "The false match rate "
            "and false non-match rate curves "
            "do not intersect each other.",
            RuntimeWarning,
        )
        eer = 1.0
        threshold = float(thresholds[0])
        return eer, Stats(fmr, fnmr, thresholds, threshold)

    t1 = t2 - 1 if diff[t2] != 0 and t2 != 0 else t2
    if fmr[t1] + fnmr[t1] <= fmr[t2] + fnmr[t2]:
        eer = (fnmr[t1] + fmr[t1]) / 2.0
        threshold = thresholds[t1]
    else:  # pragma: nocover (couldn't find a test to trigger this)
        eer = (fnmr[t2] + fmr[t2]) / 2.0
        threshold = thresholds[t2]
    eer = float(eer)
    threshold = float(threshold)
    return eer, Stats(fmr, fnmr, thresholds, threshold)


def event_error_rate(
    truth: typing.Union[str, typing.Sequence[typing.Union[str, typing.Sequence[int]]]],
    prediction: typing.Union[
        str, typing.Sequence[typing.Union[str, typing.Sequence[int]]]
    ],
) -> float:
    r"""Event error rate based on edit distance.

    The event error rate is computed by aggregating the mean edit
    distances of each (truth, prediction)-pair and averaging the
    aggregated score by the number of pairs.

    The mean edit distance of each (truth, prediction)-pair is computed
    as an average of the edit distance over the length of the longer sequence
    of the corresponding pair. By normalizing over the longer sequence the
    normalized distance is bound to [0, 1].

    Args:
        truth: ground truth classes
        prediction: predicted classes

    Returns:
        event error rate

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> event_error_rate([[0, 1]], [[0]])
        0.5
        >>> event_error_rate([[0, 1], [2]], [[0], [2]])
        0.25
        >>> event_error_rate(["lorem"], ["lorm"])
        0.2
        >>> event_error_rate(["lorem", "ipsum"], ["lorm", "ipsum"])
        0.1

    """
    truth = audeer.to_list(truth)
    prediction = audeer.to_list(prediction)

    assert_equal_length(truth, prediction)

    eer = 0.0

    for t, p in zip(truth, prediction):
        n = max(len(t), len(p))
        n = n if n > 1 else 1
        eer += edit_distance(t, p) / n

    num_samples = len(truth) if len(truth) > 1 else 1
    return eer / num_samples


def fscore_per_class(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    zero_division: float = 0,
) -> typing.Dict[str, float]:
    r"""F-score per class.

    .. math::

        \text{fscore}_k = \frac{\text{true positive}_k}
                 {\text{true positive}_k + \frac{1}{2}
                 (\text{false positive}_k + \text{false negative}_k)}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        zero_division: set the value to return when there is a zero division

    Returns:
        dictionary with label as key and F-score as value

    Examples:
        >>> fscore_per_class([0, 0], [0, 1])
        {0: 0.6666666666666666, 1: 0.0}

    """
    if labels is None:
        labels = infer_labels(truth, prediction)

    precision = precision_per_class(
        truth,
        prediction,
        labels,
        zero_division=zero_division,
    )
    recall = recall_per_class(
        truth,
        prediction,
        labels,
        zero_division=zero_division,
    )
    fscore = {}
    for label, p, r in zip(labels, precision.values(), recall.values()):
        if p * r == 0:
            fscore[label] = 0.0
        elif (p == 0.0 and np.isnan(r)) or (r == 0.0 and np.isnan(p)):
            fscore[label] = 0.0
        else:
            fscore[label] = (2 * p * r) / (p + r)
    return fscore


def linkability(
    truth: typing.Union[
        typing.Union[bool, int], typing.Sequence[typing.Union[bool, int]]
    ],
    prediction: typing.Union[
        typing.Union[bool, int, float], typing.Sequence[typing.Union[bool, int, float]]
    ],
    omega: float = 1.0,
    nbins: int = None,
) -> float:
    r"""Linkability for verification tasks.

    Let :math:`s` be the provided prediction score
    for the similarity of the tested sample.
    The clipped local linkability metric is then defined as:

    .. math::

        \text{max}(0, p(\text{mated} | s) - p(\text{non-mated} | s))

    The higher the value,
    the more likely
    that an attacker can link two mated samples.
    The global linkability metric :math:`D_\text{sys}`
    is the mean value
    over all local scores,\ :footcite:`GomezBarrero2017`
    and in the range :math:`0` and :math:`1`.

    Implementation is based on
    `code from M. Maouche`_,
    which is licensed under LGPL.

    .. footbibliography::

    .. _code from M. Maouche: https://gitlab.inria.fr/magnet/anonymization_metrics

    Args:
        truth: ground truth classes
        prediction: predicted classes or similarity scores
        omega: prior ratio
            :math:`\frac{p(\text{mated})}{p(\text{non-mated})}`
        nbins: number of bins
            of the histograms
            that estimate the distributions
            of mated and non-mated scores.
            If ``None`` it is set to
            :math:`\min(\frac{\text{len}(\text{mated})}{10}, 100)`

    Returns:
        global linkability :math:`D_\text{sys}`

    Raises:
        ValueError: if ``truth`` contains values
            different from ``1``, ``0``, ``True``, ``False``

    Examples:
        >>> np.random.seed(1)
        >>> samples = 10000
        >>> truth = [1, 0] * int(samples / 2)
        >>> prediction = []
        >>> for _ in range(int(samples / 2)):
        ...     prediction.extend([np.random.uniform(0, 0.2), np.random.uniform(0.8, 1.0)])
        >>> linkability(truth, prediction)
        0.9747999999999999
        >>> truth = [1, 0, 0, 0] * int(samples / 4)
        >>> prediction = [np.random.uniform(0, 1) for _ in range(samples)]
        >>> linkability(truth, prediction, omega=1 / 3)
        0.0

    """  # noqa: E501
    mated_scores, non_mated_scores = _matching_scores(truth, prediction)

    # Limiting the number of bins
    # (100 maximum or lower if few scores available)
    if nbins is None:
        nbins = min(int(len(mated_scores) / 10), 100)

    # Define range of scores to compute D
    bin_edges = np.linspace(
        min([min(mated_scores), min(non_mated_scores)]),
        max([max(mated_scores), max(non_mated_scores)]),
        num=nbins + 1,
        endpoint=True,
    )
    bin_centers = (bin_edges[1:] + bin_edges[:-1]) / 2

    # Compute score distributions using normalized histograms
    y1 = np.histogram(mated_scores, bins=bin_edges, density=True)[0]
    y2 = np.histogram(non_mated_scores, bins=bin_edges, density=True)[0]
    # LR = P[s|mated ]/P[s|non-mated]
    lr = np.divide(y1, y2, out=np.ones_like(y1), where=y2 != 0)
    d = 2 * (omega * lr / (1 + omega * lr)) - 1
    # Def of D
    d[omega * lr <= 1] = 0
    # Taking care of inf/NaN
    mask = [True if y2[i] == 0 and y1[i] != 0 else False for i in range(len(y1))]
    d[mask] = 1
    # Global measure using trapz numerical integration
    d_sys = np.trapz(x=bin_centers, y=d * y1)

    return float(d_sys)


def mean_absolute_error(
    truth: typing.Sequence[float],
    prediction: typing.Sequence[float],
) -> float:
    r"""Mean absolute error.

    .. math::

        \text{MAE} = \frac{1}{n} \sum^n_{i=1}
            |\text{prediction} - \text{truth}|

    Args:
        truth: ground truth values
        prediction: predicted values

    Returns:
        mean absolute error

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> mean_absolute_error([0, 0], [0, 1])
        0.5

    """
    assert_equal_length(truth, prediction)

    prediction = np.array(prediction)
    truth = np.array(truth)

    return float(np.abs(truth - prediction).mean(axis=0))


def mean_squared_error(
    truth: typing.Sequence[float],
    prediction: typing.Sequence[float],
) -> float:
    r"""Mean squared error.

    .. math::

        \text{MSE} = \frac{1}{n} \sum^n_{i=1}
            (\text{prediction} - \text{truth})^2

    Args:
        truth: ground truth values
        prediction: predicted values

    Returns:
        mean squared error

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> mean_squared_error([0, 0], [0, 1])
        0.5

    """
    assert_equal_length(truth, prediction)

    prediction = np.array(prediction)
    truth = np.array(truth)

    return float(np.square(truth - prediction).mean(axis=0))


def pearson_cc(
    truth: typing.Sequence[float],
    prediction: typing.Sequence[float],
) -> float:
    r"""Pearson correlation coefficient.

    .. math::

        \rho = \frac{\text{cov}(\text{prediction}, \text{truth})}{
        \sigma_\text{prediction}\sigma_\text{truth}}

    where :math:`\sigma` is the standard deviation,
    and :math:`\text{cov}` is the covariance.

    Args:
        truth: ground truth values
        prediction: predicted values

    Returns:
        pearson correlation coefficient :math:`\in [-1, 1]`

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> pearson_cc([0, 1, 2], [0, 1, 1])
        0.8660254037844385

    """
    assert_equal_length(truth, prediction)

    if not isinstance(truth, np.ndarray):
        truth = np.array(list(truth))
    if not isinstance(prediction, np.ndarray):
        prediction = np.array(list(prediction))

    if len(prediction) < 2 or prediction.std() == 0:
        return np.nan
    else:
        return float(np.corrcoef(prediction, truth)[0][1])


def precision_per_class(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    zero_division: float = 0,
) -> typing.Dict[str, float]:
    r"""Precision per class.

    .. math::

        \text{precision}_k = \frac{\text{true positive}_k}
                 {\text{true positive}_k + \text{false positive}_k}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        zero_division: set the value to return when there is a zero division

    Returns:
        dictionary with label as key and precision as value

    Examples:
        >>> precision_per_class([0, 0], [0, 1])
        {0: 1.0, 1: 0.0}

    """
    if labels is None:
        labels = infer_labels(truth, prediction)

    matrix = np.array(confusion_matrix(truth, prediction, labels))
    total = matrix.sum(axis=0)
    old_settings = np.seterr(invalid="ignore")
    recall = matrix.diagonal() / total
    np.seterr(**old_settings)
    recall[np.isnan(recall)] = zero_division

    return {label: float(r) for label, r in zip(labels, recall)}


def recall_per_class(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    zero_division: float = 0,
) -> typing.Dict[str, float]:
    r"""Recall per class.

    .. math::

        \text{recall}_k = \frac{\text{true positive}_k}
                 {\text{true positive}_k + \text{false negative}_k}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        zero_division: set the value to return when there is a zero division

    Returns:
        dictionary with label as key and recall as value

    Examples:
        >>> recall_per_class([0, 0], [0, 1])
        {0: 0.5, 1: 0.0}

    """
    if labels is None:
        labels = infer_labels(truth, prediction)

    matrix = np.array(confusion_matrix(truth, prediction, labels))
    total = matrix.sum(axis=1)
    old_settings = np.seterr(invalid="ignore")
    recall = matrix.diagonal() / total
    np.seterr(**old_settings)
    recall[np.isnan(recall)] = zero_division

    return {label: float(r) for label, r in zip(labels, recall)}


def unweighted_average_bias(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    protected_variable: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    subgroups: typing.Sequence[typing.Any] = None,
    metric: typing.Callable[
        [
            typing.Sequence[typing.Any],
            typing.Sequence[typing.Any],
            typing.Optional[typing.Sequence[str]],
        ],
        typing.Dict[str, float],
    ] = fscore_per_class,
    reduction: typing.Callable[
        [
            typing.Sequence[float],
        ],
        float,
    ] = np.std,
) -> float:
    r"""Unweighted average bias of protected variable.

    The bias is measured in terms of *equalized odds* which requires
    the classifier to have identical performance for all classes independent
    of a protected variable such as race. The performance of the classifier
    for its different classes can be assessed with standard metrics
    such as *recall* or *precision*. The difference in performance, denoted
    as score divergence, can be computed in different ways, as well.
    For two subgroups the (absolute) difference serves as a standard choice.
    For more than two subgroups the score divergence could be estimated by
    the standard deviation of the scores.

    Note:
        If for a class less than two subgroups exhibit a performance score,
        the corresponding class is ignored in the bias computation.
        This occurs if there is no class sample for a subgroup,
        e.g. no negative (class label) female (subgroup of sex).

    Args:
        truth: ground truth classes
        prediction: predicted classes
        protected_variable: manifestations of protected variable such as
            subgroups "male" and "female" of variable "sex"
        labels: included labels in preferred ordering.
            The bias is computed only on the specified labels.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        subgroups: included subgroups in preferred ordering.
            The direction of the bias is determined by the ordering of the
            subgroups.
            Besides, the bias is computed only on the specified subgroups.
            If no subgroups are supplied, they will be inferred from
            :math:`\text{protected\_variable}` and ordered alphanumerically.
        metric: metric which equalized odds are measured with.
            Typical choices are: :func:`audmetric.recall_per_class`,
            :func:`audmetric.precision_per_class` or
            :func:`audmetric.fscore_per_class`
        reduction: specifies the reduction operation to measure the divergence
            between the scores of the subgroups of the protected variable
            for each class. Typical choices are:
            difference or absolute difference between scores for two subgroups
            and standard deviation of scores for more than two subgroups.

    Returns:
        unweighted average bias

    Raises:
        ValueError: if ``truth``, ``prediction`` and ``protected_variable``
            have different lengths
        ValueError: if ``subgroups`` contains values not contained in
            ``protected_variable``

    Examples:
        >>> unweighted_average_bias([1, 1], [1, 0], ["male", "female"])
        0.5
        >>> unweighted_average_bias(
        ...     [1, 1],
        ...     [1, 0],
        ...     ["male", "female"],
        ...     subgroups=["female", "male"],
        ...     reduction=lambda x: x[0] - x[1],
        ... )
        -1.0
        >>> unweighted_average_bias([0, 1], [1, 0], ["male", "female"], metric=recall_per_class)
        nan
        >>> unweighted_average_bias(
        ...     [0, 0, 0, 0],
        ...     [1, 1, 0, 0],
        ...     ["a", "b", "c", "d"],
        ...     metric=recall_per_class,
        ... )
        0.5

    """  # noqa: E501
    if labels is None:
        labels = infer_labels(truth, prediction)

    if not len(truth) == len(prediction) == len(protected_variable):
        raise ValueError(
            f"'truth', 'prediction' and 'protected_variable' should have "
            f"same lengths, but received '{len(truth)}', '{len(prediction)}' "
            f"and '{len(protected_variable)}'"
        )

    if subgroups is None:
        subgroups = sorted(set(protected_variable))

    scores = scores_per_subgroup_and_class(
        truth=truth,
        prediction=prediction,
        protected_variable=protected_variable,
        labels=labels,
        subgroups=subgroups,
        metric=metric,
        zero_division=np.nan,
    )

    bias = 0.0
    denominator = 0

    for label in labels:
        scores_subgroup = [
            scores[subgroup][label]
            for subgroup in subgroups
            if label in scores[subgroup] and not np.isnan(scores[subgroup][label])
        ]
        # compute score divergence only where more than 1 score per class
        if len(scores_subgroup) > 1:
            bias += reduction(scores_subgroup)
            denominator += 1

    if denominator == 0:
        return np.nan

    return float(bias / denominator)


def unweighted_average_fscore(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    zero_division: float = 0,
) -> float:
    r"""Unweighted average F-score.

    .. math::

        \text{UAF} = \frac{1}{K} \sum^K_{k=1}
            \frac{\text{true positive}_k}
                 {\text{true positive}_k + \frac{1}{2}
                 (\text{false positive}_k + \text{false negative}_k)}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        zero_division: set the value to return when there is a zero division

    Returns:
        unweighted average precision

    Examples:
        >>> unweighted_average_fscore([0, 0], [0, 1])
        0.3333333333333333

    """
    fscore = fscore_per_class(
        truth,
        prediction,
        labels,
        zero_division=zero_division,
    )
    fscore = np.array(list(fscore.values()))
    return float(fscore.mean())


def unweighted_average_precision(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    zero_division: float = 0,
) -> float:
    r"""Unweighted average precision.

    .. math::

        \text{UAP} = \frac{1}{K} \sum^K_{k=1}
            \frac{\text{true positive}_k}
                 {\text{true positive}_k + \text{false positive}_k}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        zero_division: set the value to return when there is a zero division

    Returns:
        unweighted average precision

    Examples:
        >>> unweighted_average_precision([0, 0], [0, 1])
        0.5

    """
    precision = precision_per_class(
        truth,
        prediction,
        labels,
        zero_division=zero_division,
    )
    precision = np.array(list(precision.values()))
    return float(precision.mean())


def unweighted_average_recall(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    labels: typing.Sequence[typing.Any] = None,
    *,
    zero_division: float = 0,
) -> float:
    r"""Unweighted average recall.

    .. math::

        \text{UAR} = \frac{1}{K} \sum^K_{k=1}
            \frac{\text{true positive}_k}
                 {\text{true positive}_k + \text{false negative}_k}

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.
        zero_division: set the value to return when there is a zero division

    Returns:
        unweighted average recall

    Examples:
        >>> unweighted_average_recall([0, 0], [0, 1])
        0.25

    """
    recall = recall_per_class(
        truth,
        prediction,
        labels,
        zero_division=zero_division,
    )
    recall = np.array(list(recall.values()))
    return float(recall.mean())


def weighted_confusion_error(
    truth: typing.Sequence[typing.Any],
    prediction: typing.Sequence[typing.Any],
    weights: typing.Sequence[typing.Sequence[typing.Union[int, float]]],
    labels: typing.Sequence[typing.Any] = None,
) -> float:
    r"""Weighted confusion error.

    Computes the normalized confusion matrix, applies given weights to each
    cell and sums them up. Weights are expected as positive numbers and
    will be normalized by the sum of all weights. The higher the weight,
    the more costly will be the error. A weight of 0 means that the cell
    is not taken into account for the error, this is usually the case for the
    diagonal as it holds correctly classified samples.

    Args:
        truth: ground truth values/classes
        prediction: predicted values/classes
        weights: weights applied to the confusion matrix.
            Expected as a list of lists in the following form
            (r=row, c=column):
            ``[[w_r0_c0, ..., w_r0_cN], ..., [w_rN_c0, ..., w_rN_cN]]``
        labels: included labels in preferred ordering.
            If no labels are supplied,
            they will be inferred from
            :math:`\{\text{prediction}, \text{truth}\}`
            and ordered alphabetically.

    Returns:
        weighted confusion error

    Examples:
        >>> truth = [0, 1, 2]
        >>> prediction = [0, 2, 0]
        >>> # penalize only errors > 1
        >>> weights = [[0, 0, 1], [0, 0, 0], [1, 0, 0]]
        >>> weighted_confusion_error(truth, prediction, weights)
        0.5

    """
    weights = weights / np.sum(weights)
    cm = confusion_matrix(truth, prediction, labels, normalize=True)
    cm = np.array(cm)

    if not cm.shape == weights.shape:
        raise ValueError(
            "Shape of weights "
            f"{weights.shape} "
            "does not match shape of confusion matrix "
            f"{cm.shape}."
        )

    weighted_cm = cm * weights
    return float(np.sum(weighted_cm))


def word_error_rate(
    truth: typing.Sequence[typing.Sequence[str]],
    prediction: typing.Sequence[typing.Sequence[str]],
) -> float:
    r"""Word error rate based on edit distance.

    Args:
        truth: ground truth strings
        prediction: predicted strings

    Returns:
        word error rate

    Raises:
        ValueError: if ``truth`` and ``prediction`` differ in length

    Examples:
        >>> truth = [["lorem", "ipsum"], ["north", "wind", "and", "sun"]]
        >>> prediction = [["lorm", "ipsum"], ["north", "wind"]]
        >>> word_error_rate(truth, prediction)
        0.5

    """
    assert_equal_length(truth, prediction)

    wer = 0.0

    for t, p in zip(truth, prediction):
        # map words to ints
        unique_words = set(t).union(set(p))
        map = {k: v for k, v in zip(unique_words, range(len(unique_words)))}
        t = [map[i] for i in t]
        p = [map[i] for i in p]

        n = max(len(t), len(p))
        n = n if n > 1 else 1
        wer += edit_distance(t, p) / n

    num_samples = len(truth) if len(truth) > 1 else 1
    return float(wer / num_samples)


def _matching_scores(
    truth: typing.Union[
        typing.Union[bool, int], typing.Sequence[typing.Union[bool, int]]
    ],
    prediction: typing.Union[
        typing.Union[bool, int, float], typing.Sequence[typing.Union[bool, int, float]]
    ],
) -> typing.Tuple[np.ndarray, np.ndarray]:
    r"""Mated and non-mated scores for verification tasks.

    For verification task,
    predictions are usually separated
    in all predictions belonging
    to the matching examples,
    and all other predictions.
    The first are called mated scores
    or genuine matching scores,
    the second non-mated scores
    or impostor matching scores.

    For example,
    in a speaker verification task
    the mated scores are all similarity values
    that belong to the matching speaker.

    Args:
        truth: ground truth classes
        prediction: predicted classes or similarity scores

    Returns:
        * mated scores
        * non-mated scores

    Raises:
        ValueError: if ``truth`` contains values
            different from ``1, 0, True, False``

    Examples:
        >>> truth = [1, 0]
        >>> prediction = [0.9, 0.1]
        >>> _matching_scores(truth, prediction)
        (array([0.9]), array([0.1]))

    """
    truth = np.array(truth)

    allowed_truth_values = set([1, 0, True, False])
    if not set(truth).issubset(allowed_truth_values):
        raise ValueError(
            "'truth' is only allowed to contain "
            "[1, 0, True, False], "
            'yours contains:\n'
            f"[{', '.join([str(t) for t in set(truth)])}]"
        )

    truth = truth.astype(bool)
    prediction = np.array(prediction).astype(np.float64)

    # Predictions for all matching examples
    # (truth is 1 or True)
    # In literature these are called
    # "genuine matching scores"
    # or "mated scores"
    mated_scores = prediction[truth]
    # Predictions for all non-matching examples
    # (truth is 0 or False)
    # In literature these are called
    # "impostor matching scores"
    # or "non-mated scores"
    non_mated_scores = prediction[~truth]

    return mated_scores, non_mated_scores
